from langchain_openai import ChatOpenAI
from langchain_core.prompts import ChatPromptTemplate
from pydantic import BaseModel

from extractor.models import ExtractedNotes, Note, NoteType, SampleExtraction
from extractor.prompt import get_extractor_prompt_template
from extractor.utils import get_messages_from_example, get_sample_as_input


def get_openai_extractor(prompt_template: ChatPromptTemplate, schema_class: BaseModel):
    llm = ChatOpenAI(model="gpt-4o", temperature=0)

    return prompt_template | llm.with_structured_output(schema=schema_class)


def perform_extraction(text: str):
    extractor = get_openai_extractor(
        get_extractor_prompt_template(),
        ExtractedNotes,
    )

    examples = []
    for example in [
        SampleExtraction(
            input=get_sample_as_input("no-lessons"),
            extracted_notes=ExtractedNotes(
                notes=[
                    Note(
                        text="Price behavior during a mania",
                        type=NoteType.KEY_THEME,
                    ),
                    Note(
                        text="Corruption and its revelation during economic cycles",
                        type=NoteType.KEY_THEME,
                    ),
                    Note(
                        text="Monitor for signs of continual price increases as an indicator of a potential mania.",
                        type=NoteType.ACTION_ITEM,
                    ),
                ]
            ),
        ),
        SampleExtraction(
            input=get_sample_as_input("no-key-themes"),
            extracted_notes=ExtractedNotes(
                notes=[
                    Note(
                        text="Focusing on high-priced products can lead to significant referral commissions.",
                        type=NoteType.LESSON,
                    ),
                    Note(
                        text="Consistently posting on a blog or social media can build a steady following and generate referrals.",
                        type=NoteType.LESSON,
                    ),
                    Note(
                        text="Identify and promote high-priced products in your affiliate marketing efforts.",
                        type=NoteType.ACTION_ITEM,
                    ),
                    Note(
                        text="Develop and maintain a consistent posting schedule for your blog or social media accounts to build and sustain a follower base.",
                        type=NoteType.ACTION_ITEM,
                    ),
                ]
            ),
        ),
        SampleExtraction(
            input=get_sample_as_input("no-action-items"),
            extracted_notes=ExtractedNotes(
                notes=[
                    Note(
                        text="Opinions outside one's expertise",
                        type=NoteType.KEY_THEME,
                    ),
                    Note(
                        text="Technology and innovation to harness environmental power",
                        type=NoteType.KEY_THEME,
                    ),
                    Note(
                        text="Forceful displacement of people for development",
                        type=NoteType.KEY_THEME,
                    ),
                    Note(
                        text="Self-developed talent without formal education can lead to misguided or harmful views.",
                        type=NoteType.LESSON,
                    ),
                    Note(
                        text="Popularity can tempt individuals to opine on topics outside their expertise.",
                        type=NoteType.LESSON,
                    ),
                    Note(
                        text="Technological innovations, such as dams, can have significant environmental impacts.",
                        type=NoteType.LESSON,
                    ),
                    Note(
                        text="Development projects often involve the displacement of people, echoing past and present experiences.",
                        type=NoteType.LESSON,
                    ),
                ]
            ),
        ),
    ]:
        examples.extend(get_messages_from_example(example))

    return extractor.invoke(
        {
            "text": text,
            "examples": examples,
        }
    )



import uuid
from typing import Optional, TypedDict
from langchain_core.prompts import ChatPromptTemplate, MessagesPlaceholder
from langchain_core.messages import BaseMessage, AIMessage, HumanMessage, ToolMessage
from langchain_core.pydantic_v1 import BaseModel, Field
from langchain_openai import ChatOpenAI
from langchain_mistralai import ChatMistralAI
import dotenv

dotenv.load_dotenv()


class Person(BaseModel):
    """Information about a person."""

    # NOTE: that doc string is actually sent to the LLM, so I need to use it wisely to describe what the model represents.

    # NOTE: for the fields:
    # - typing a string as optional allows the model to decline to extract it
    # - the description field is used by the model, so adding it can improve extraction results
    name: Optional[str] = Field(default=None, description="The name of the person")
    hair_color: Optional[str] = Field(
        default=None, description="The color of the person's hair, if known"
    )
    height_in_meters: Optional[str] = Field(
        default=None, description="Height measured in meters"
    )


prompt = ChatPromptTemplate.from_messages(
    [
        (
            "system",
            "You are an expert extraction algorithm. "
            "Only extract relevant information from the text. "
            "If you do not know the value of an attribute you're asked to extract, "
            "return null for the attribute's value.",
        ),
        MessagesPlaceholder("examples"),
        ("human", "{text}"),
    ]
)


def get_open_ai_runnable(schema_klass):
    llm = ChatOpenAI(model="gpt-4o", temperature=0)

    return prompt | llm.with_structured_output(schema=schema_klass)


def get_mistral_ai_runnable(schema_klass):
    llm = ChatMistralAI(model="mistral-large-latest", temperature=0)

    return prompt | llm.with_structured_output(schema=schema_klass)


text = "Alan Smith is 6 feet tall and has blond hair."
# res = get_open_ai_runnable(schema_klass=Person).invoke({"text": text})
# res = get_open_ai_runnable(schema_klass=Person).invoke({"text": text})


class Data(BaseModel):
    """Extracted data about people"""

    people: list[Person]


# runnable = get_open_ai_runnable(schema_klass=Data).invoke(
#     {
#         "text": "My name is Jeff, my hair is black and i am 6 feet tall. Anna has the same color hair as me."
#     }
# )


class Example(TypedDict):
    """
    A representation of an example consisting of the text input, expected tool calls, and optionally the tool call results.

    For extraction tasks, the tool calls are represented as instances of the Pydantic BaseModel and the tool call results can be simply sentences telling the model that it called the tool correctly.
    """

    input: str
    tool_calls: list[BaseModel]
    tool_call_responses: Optional[list[str]]


def tool_example_to_messages(example: Example) -> list[BaseMessage]:
    """
    Converts an example into messages that can be fed into the model.

    For each example this function will output 3 types of messages:
    (1) HumanMessage: input text from which content should be extracted
    (2) AIMessage: sample response from the LLM that will contain tool call definition(s)
    (3) ToolMessage: sample response back to the LLM to confirm that its tool call definitions were correct
    """
    # first specify the human message
    messages: list[BaseMessage] = [HumanMessage(content=example["input"])]

    # then prepare the tool call definitions for the AI message
    openai_tool_calls = []
    for tool_call in example["tool_calls"]:
        openai_tool_calls.append(
            {
                "id": str(uuid.uuid4()),
                "type": "function",
                "function": {
                    # for now the name of the function is the name of the pydantic
                    # model class that we want to see in the output
                    "name": tool_call.__class__.__name__,
                    "arguments": tool_call.json(),
                },
            }
        )
    # then specify the AI message using those definitions
    messages.append(
        AIMessage(content="", additional_kwargs={"tool_calls": openai_tool_calls})
    )
    # then specify the tool message
    tool_call_responses = example.get("tool_call_responses") or [
        "You have correctly called this tool."
    ] * len(openai_tool_calls)
    # each tool call response must be mapped to the ID of its corresponding tool call definition
    messages.extend(
        ToolMessage(content=response, tool_call_id=tool_call["id"])
        for response, tool_call in zip(tool_call_responses, openai_tool_calls)
    )

    return messages


examples = [
    (
        "The ocean is vast and blue. It's more than 20,000 feet deep. There are many fish in it.",
        Person(name=None, hair_color=None, height_in_meters=None),
    ),
    (
        "Fiona traveled far from France to Spain.",
        Person(name="Fiona", hair_color=None, height_in_meters=None),
    ),
]

messages = []
for text, extracted_person in examples:
    messages.extend(
        tool_example_to_messages(Example(input=text, tool_calls=[extracted_person]))
    )

res = prompt.invoke({"text": "this is some text", "examples": messages})


def get_open_ai_extractor(schema_klass):
    llm = ChatOpenAI(model="gpt-4-0125-preview", temperature=0)

    return prompt | llm.with_structured_output(schema=schema_klass, include_raw=False)


print("========= without examples =========")
for _i in range(5):
    text = "The solar system is large, but earth has only 1 moon."
    print(get_open_ai_extractor(Person).invoke({"text": text, "examples": []}))

print("========= with examples =========")
for _i in range(5):
    text = "The solar system is large, but earth has only 1 moon."
    print(get_open_ai_extractor(Person).invoke({"text": text, "examples": messages}))



from uuid import uuid4
from typing import TypedDict, Literal
from langchain_core.messages import HumanMessage, AIMessage, ToolMessage, BaseMessage
from extractor.models import SampleExtraction


class OpenAIFunctionCallFunctionFieldDef(TypedDict):
    name: str
    arguments: str


class OpenAIFunctionCallDef(TypedDict):
    id: str
    type: Literal["function"]
    function: OpenAIFunctionCallFunctionFieldDef


def get_messages_from_example(example: SampleExtraction) -> list[BaseMessage]:
    """
    Converts an example note into a list of messages that can be fed into an LLM.

    The conversion is as follows:
    - "input": transformed into a single `HumanMessage` object
    - "extracted_notes": transformed into a list of dictionaries that define tool calls for an `AIMessage` object
    """
    messages = [HumanMessage(content=example["input"])]

    tool_calls: list[OpenAIFunctionCallDef] = []
    for extracted_note in example["extracted_notes"].notes:
        tool_calls.append(
            {
                "id": str(uuid4()),
                "type": "function",
                "function": {
                    # for now langchain accepts the name of the pydantic model as the name of the function
                    "name": extracted_note.__class__.__name__,
                    "arguments": extracted_note.json(),
                },
            }
        )

    messages.append(AIMessage(content="", additional_kwargs={"tool_calls": tool_calls}))

    # append ToolMessages to make sure the LLM gets feedback on the result of the function calls
    tool_message_texts = ["You have called the function correctly."] * len(tool_calls)
    messages.extend(
        ToolMessage(content=msg_text, tool_call_id=tool_call["id"])
        for msg_text, tool_call in zip(tool_message_texts, tool_calls)
    )

    return messages


def get_sample_as_input(file_name: str) -> str:
    with open((f"extractor/examples/{file_name}.txt"), "r") as file:
        return file.read()



from dotenv import load_dotenv
from langchain_core.prompts import ChatPromptTemplate
from langchain_openai import ChatOpenAI


load_dotenv()


prompt = ChatPromptTemplate.from_messages(
    [
        (
            "system",
            "You're an expert summarizer. "
            "Give me 3 takeaways from the text I provide, formatted as bullet points.",
        ),
        ("human", "{text}"),
    ]
)


def get_open_ai_summarizer():
    llm = ChatOpenAI(model="gpt-4o", temperature=0)

    return prompt | llm


text = "Manchester United Football Club, commonly referred to as Man United (often stylised as Man Utd), or simply United, is a professional football club based in Old Trafford, Greater Manchester, England. The club competes in the Premier League, the top tier of English football. Nicknamed the Red Devils, they were founded as Newton Heath LYR Football Club in 1878, but changed their name to Manchester United in 1902. After a spell playing in Clayton, Manchester, the club moved to their current stadium, Old Trafford, in 1910."

print(get_open_ai_summarizer().invoke({"text": text}).content)



from langchain_openai import ChatOpenAI
import dotenv

from summarizer.prompting.prompt import get_summarizer_prompt


dotenv.load_dotenv()


def get_open_ai_summarizer():
    return get_summarizer_prompt() | ChatOpenAI(model="gpt-4o", temperature=0)


def perform_summarization(text: str):
    return get_open_ai_summarizer().invoke({"text": text})



from langchain_core.prompts import ChatPromptTemplate, MessagesPlaceholder


def get_extractor_prompt_template():
    prompt_template = ChatPromptTemplate.from_messages(
        [
            (
                "system",
                "You are an expert extraction algorithm. "
                "You are to extract key themes, lessons, and potential action items from the text. "
                "If you do not know the value of an attribute you're asked to extract, "
                "return null for the attribute's value.",
            ),
            MessagesPlaceholder("examples"),
            ("human", "{text}"),
        ]
    )

    return prompt_template



from setuptools import setup, find_packages


# Function to read the requirements.txt file
def read_requirements():
    with open("requirements.txt") as f:
        return f.read().splitlines()


setup(
    name="ai-notes-summaries",
    version="0.1.2",
    packages=find_packages(),
    install_requires=read_requirements(),
    url="https://github.com/CodeWithOz/ai-notes-summaries",
    author="Uche Ozoemena",
    author_email="codewithoz@outlook.com",
    description="An AI-powered tool that extracts structured notes from a given piece of text.",
)



from langchain_core.prompts import ChatPromptTemplate


def get_summarizer_prompt():
    with open((f"summarizer/prompting/summarizer.txt"), "r") as file:
        prompt = file.read()

    return ChatPromptTemplate.from_messages(
        [
            (
                "system",
                prompt,
            ),
            ("human", "{text}"),
        ]
    )



from typing import Optional, TypedDict
from enum import IntEnum
from langchain_core.pydantic_v1 import BaseModel, Field


class NoteType(IntEnum):
    """The type of note. Currently only supports 1, 2, and 3."""

    KEY_THEME = 1
    LESSON = 2
    ACTION_ITEM = 3


class Note(BaseModel):
    """A description and/or summary of an idea/concept."""

    text: Optional[str] = Field(
        default=None,
        description="Text that describes and/or summarizes the idea/concept.",
    )
    type: Optional[NoteType] = Field(
        default=None,
        description="The type of note. Currently only supports 1, 2, and 3 as defined in the NoteType enum. 1 is a key theme, 2 is a lesson, and 3 is a potential action item.",
    )


class ExtractedNotes(BaseModel):
    """A collection of `Note` objects."""

    notes: list[Note]


class SampleExtraction(TypedDict):
    # the text on which extraction should be performed
    input: str

    # the notes that should be extracted
    extracted_notes: ExtractedNotes
